/*
 * Copyright 2012 <James.Bottomley@HansenPartnership.com>
 *
 * see COPYING file
 */
#include <stdint.h>
#define __STDC_VERSION__ 199901L
#include <efi.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <unistd.h>
#include <wchar.h>

#include <PeImage.h>		/* for ALIGN_VALUE */
#include <sha256.h>
#include <efiauthenticated.h>
#include <guid.h>
#include <version.h>

static void
usage(const char *progname)
{
	printf("Usage: %s [-b efi-binary [efi-binary ...]] "
	       "[-d hex-sha256-value [hex-sha256-value ...]] efi-signature-list\n",
	   progname);
}

static void
help(const char *progname)
{
	usage(progname);
	printf("Produce an EFI Signature List file containing the one or more sha256"
	       " hashes\n\nOptions:\n"
	       "\t-b  efi-binary [efi-binary ...]             Calculate sha256 value from one or more EFI binaries\n"
	       "\t-d  hex-sha256-value [hex-sha256-value ...] Use provided sha256 values\n"
	       );
}

/*
 * Convert one ASCII hex character to integer.
 * If cannot be converted, set to -1
 */
#define HEX2INT(c, t) do { \
	if ((c) >= 'A' && (c) <= 'F') { \
		(t) =  (c) - 'A' + 10; \
	} else if ((c) >= 'a' && (c) <= 'f') { \
		(t) = (c) - 'a' + 10; \
	} else if ((c) >= '0' && (c) <= '9') { \
		(t) = (c) - '0'; \
	} else { \
		/* Error input */ \
		(t) = -1; \
	} \
} while (0);

int
main(int argc, char *argv[])
{
	void *efifile;
	const char *progname = argv[0];
	int i, need_hashing;

	/* By default, we assume -b mode */
	need_hashing = 1;

	while (argc > 1) {
		if (strcmp("--version", argv[1]) == 0) {
			version(progname, NULL);
			exit(0);
		} else if (strcmp("--help", argv[1]) == 0) {
			help(progname);
			exit(0);
		} else if (strcmp("-b", argv[1]) == 0) {
			need_hashing = 1;
			argc -= 1;
			argv += 1;
		} else if (strcmp("-d", argv[1]) == 0) {
			need_hashing = 0;
			argc -= 1;
			argv += 1;
		} else {
			break;
		}
	}

	if (argc < 3) {
		usage(progname);
		exit(1);
	}

	int hashes = argc - 2;
	UINT8 hash[hashes][SHA256_DIGEST_SIZE];

	memset(hash, 0, sizeof(hash));

	for (i = 0; i < hashes; i++) {
		int j;
		struct stat st;
		EFI_STATUS status;

		if (need_hashing) {
			int fdefifile = open(argv[i + 1], O_RDONLY);
			if (fdefifile == -1) {
				fprintf(stderr, "failed to open file %s: ", argv[1]);
				perror("");
				exit(1);
			}
			fstat(fdefifile, &st);
			efifile = malloc(ALIGN_VALUE(st.st_size, 4096));
			memset(efifile, 0, ALIGN_VALUE(st.st_size, 4096));
			read(fdefifile, efifile, st.st_size);
			close(fdefifile);
			status = sha256_get_pecoff_digest_mem(efifile, st.st_size,
			    hash[i]);
			if (status != EFI_SUCCESS) {
				printf("Failed to get hash of %s: %ld\n", argv[i+1],
				    status);
				exit(1);
			}
		} else {
			const char *input = argv[i + 1];
			int t, valid = 1;
			size_t len;

			len = strlen(input);

			/* allow 0x prefixed hashes */
			if (len > 2 && strncasecmp(input, "0x", 2) == 0) {
				input += 2;
				len -= 2;
			}

			if (len != SHA256_DIGEST_SIZE * 2) {
				printf("Improperly formatted sha256 hash "
				    "(invalid length have:%d want:64): %s\n",
				    (int)len, input);
				continue;
			}
			for (j = 0; j < SHA256_DIGEST_SIZE; j++) {
				HEX2INT(input[j * 2], t);
				if (t == -1) {
					valid = 0;
					break;
				}
				hash[i][j] = t << 4;

				HEX2INT(input[j * 2 + 1], t);
				if (t == -1) {
					valid = 0;
					break;
				}
				hash[i][j] |= t;
			}
			len = strlen("Error:");
			if (valid == 0) {
				printf("Improperly formatted sha256 hash:\n%*s%s\n",
				    (int)len, "", input);
				printf("%-*s%s\n", j*2 + (int)len, "Error:", "^^");
				exit(1);
			}
		}
		printf("HASH IS ");
		for (j = 0; j < SHA256_DIGEST_SIZE; j++) {
			printf("%02x", hash[i][j]);
		}
		printf("\n");
	}
	UINT8 sig[sizeof(EFI_SIGNATURE_LIST) + (sizeof(EFI_SIGNATURE_DATA) - 1 + SHA256_DIGEST_SIZE) * hashes];

	EFI_SIGNATURE_LIST *l = (void *)sig;

	memset(sig, 0, sizeof(sig));
	l->SignatureType = EFI_CERT_SHA256_GUID;
	l->SignatureListSize = sizeof(sig);
	l->SignatureSize = 16 +32; /* UEFI defined */
	for (i = 0; i < hashes; i++) {
		EFI_SIGNATURE_DATA *d = (void *)sig + sizeof(EFI_SIGNATURE_LIST) + l->SignatureSize * i;
		d->SignatureOwner = MOK_OWNER;
		memcpy(&d->SignatureData, hash[i], sizeof(hash[i]));
	}

	int fdoutfile = open(argv[hashes + 1], O_CREAT|O_WRONLY|O_TRUNC, S_IWUSR|S_IRUSR);
	if (fdoutfile == -1) {
		fprintf(stderr, "failed to open %s: ", argv[hashes + 1]);
		perror("");
		exit(1);
	}
	write(fdoutfile, sig, sizeof(sig));
	close(fdoutfile);
	return 0;
}
